# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import hashlib
import itertools
import logging
import random
from collections.abc import Callable, Iterator
from typing import TypeVar

import attrs
import lance
import more_itertools
import pyarrow as pa

from geneva.checkpoint import (
    CheckpointStore,
)
from geneva.debug.logger import ErrorLogger, NoOpErrorLogger
from geneva.query import Scan
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


@attrs.define(order=True)
class ReadTask:
    uri: str
    columns: list[str]
    frag_id: int
    offset: int
    limit: int

    filter: str | None = None

    def to_batches(self, *, batch_size=32) -> Iterator[pa.RecordBatch]:
        uri_parts = self.uri.split("/")
        name = ".".join(uri_parts[-1].split(".")[:-1])
        db = "/".join(uri_parts[:-1])
        scan = (
            Scan.from_uri(db, name)
            .with_columns(self.columns)
            .with_fragments([self.frag_id])
            .with_filter(self.filter)
            .with_offset(self.offset)
            .with_limit(self.limit)
        )
        yield from scan.to_batches(batch_size)

    def checkpoint_key(self) -> str:
        hasher = hashlib.md5()
        hasher.update(
            f"{self.uri}:{self.frag_id}:{self.offset}:{self.limit}:{self.filter}".encode(),
        )
        return hasher.hexdigest()


@attrs.define
class LanceRecordBatchUDFApplier:
    udfs: dict[str, UDF] = attrs.field()
    checkpoint_store: CheckpointStore = attrs.field()
    error_logger: ErrorLogger = attrs.field(default=NoOpErrorLogger())

    batch_size: int = 32

    @property
    def output_schema(self) -> pa.Schema:
        return pa.schema(
            [pa.field(name, fn.data_type) for name, fn in self.udfs.items()],
        )

    def _run(self, task: ReadTask) -> dict[str, str]:
        data_key = task.checkpoint_key()
        _LOG.debug("Running task %s", task)
        # track the batch sequence number so we can checkpoint any errors
        # when reproducing locally we can seek to the erroring batch quickly

        # prepare the schema
        fields = []
        for name, fn in self.udfs.items():
            fields.append(pa.field(name, fn.data_type, metadata=fn.field_metadata))

        res = {}
        batch = None
        for name, fn in self.udfs.items():
            checkpoint_key = f"{data_key}:{fn.checkpoint_key}"
            if checkpoint_key in self.checkpoint_store:
                _LOG.info("Using cached result for %s", checkpoint_key)
                res[name] = checkpoint_key
                continue
            arrs = []
            # PERF203 -- don't try-except inside the loop
            # so I had to move the loop inside the try-except
            # and need some loop state tracking for error logggin
            seq = 0
            # TODO: add caching for the input data
            try:
                for _seq, batch in enumerate(
                    task.to_batches(batch_size=self.batch_size)
                ):
                    seq = _seq
                    arrs.append(fn(batch))
            except Exception as e:
                self.error_logger.log_error(e, task, fn, seq)
                raise e

            arr = pa.concat_arrays(arrs)
            self.checkpoint_store[checkpoint_key] = pa.RecordBatch.from_pydict(
                {"data": arr},
                schema=pa.schema([pa.field("data", fn.data_type)]),
            )
            res[name] = checkpoint_key

        return res

    def run(self, task: ReadTask) -> dict[str, str]:
        try:
            return self._run(task)
        except Exception as e:
            logging.exception("Error running task %s: %s", task, e)
            raise RuntimeError(f"Error running task {task}") from e

    def status(self, task: ReadTask) -> dict[str, str]:
        data_key = task.checkpoint_key()
        return {
            name: f"{data_key}:{fn.checkpoint_key}" in self.checkpoint_store
            for name, fn in self.udfs.items()
        }  # type: ignore


def _plan_read(
    uri: str,
    columns: list[str] | None = None,
    *,
    read_version: int | None = None,
    batch_size: int = 512,
    filter: str | None = None,  # noqa: A002
    num_frags: int | None = None,
) -> Iterator[ReadTask]:
    """Make Plan for Reading Data from a Dataset"""
    if columns is None:
        columns = []
    dataset = lance.dataset(uri)
    if read_version is not None:
        dataset = dataset.checkout_version(read_version)

    for idx, frag in enumerate(dataset.get_fragments()):
        if num_frags is not None and idx >= num_frags:
            break
        frag_rows = frag.count_rows(filter=filter)
        for offset in range(0, frag_rows, batch_size):
            limit = min(batch_size, frag_rows - offset)
            yield ReadTask(
                uri=uri,
                columns=columns,
                frag_id=frag.fragment_id,
                offset=offset,
                limit=limit,
                filter=filter,
            )


@attrs.define
class _LanceReadPlanIterator(Iterator[ReadTask]):
    it: Iterator[ReadTask]
    total: int

    def __iter__(self) -> Iterator[ReadTask]:
        return self

    def __next__(self) -> ReadTask:
        return next(self.it)

    def __len__(self) -> int:
        return self.total


def _num_tasks(
    *,
    uri: str,
    read_version: int | None = None,
    batch_size: int = 512,
) -> int:
    return sum(
        -(-frag.count_rows() // batch_size)
        for frag in (lance.dataset(uri, version=read_version)).get_fragments()
    )


T = TypeVar("T")


def _buffered_shuffle(it: Iterator[T], buffer_size: int) -> Iterator[T]:
    """Shuffle an iterator using a buffer of size buffer_size
    not perfectly random, but good enough for spreading out IO
    """
    # Initialize the buffer with the first buffer_size items from the iterator
    buffer = []
    # Fill the buffer with up to buffer_size items initially
    try:
        for _ in range(buffer_size):
            item = next(it)
            buffer.append(item)
    except StopIteration:
        pass

    while True:
        # Select a random item from the buffer
        index = random.randint(0, len(buffer) - 1)
        item = buffer[index]

        # Try to replace the selected item with a new one from the iterator
        try:
            next_item = next(it)
            buffer[index] = next_item
            # Yield the item AFTER replacing it in the buffer
            # this way the buffer is always contiguous so we can
            # simply yield the buffer at the end
            yield item
        except StopIteration:
            yield from buffer
            break


R = TypeVar("R")


def diversity_aware_shuffle(
    it: Iterator[T],
    key: Callable[[T], R],
    *,
    diversity_goal: int = 4,
    buffer_size: int = 1024,
) -> Iterator[T]:
    """A shuffle iterator that is aware of the diversity of the data
    being shuffled. The key function should return a value that is
    is used to determine the diversity of the data. The diversity_goal
    is the number of unique values that should be in the buffer at any
    given time. if the buffer is full, the items is yielded in a round-robin
    fashion. This is useful for shuffling tasks that are diverse, but

    This algorithm is bounded in memory by the buffer_size, so it is reasonably
    efficient for large datasets.
    """

    # NOTE: this is similar to itertools.groupby, but with a buffering limit

    buffer: dict[R, list[T]] = {}
    buffer_total_size = 0

    peekable_it = more_itertools.peekable(it)

    def _maybe_consume_from_iter() -> bool:
        nonlocal buffer_total_size
        item = peekable_it.peek(default=None)
        if item is None:
            return False
        key_val = key(item)
        if key_val not in buffer and len(buffer) < diversity_goal:
            buffer[key_val] = []
        else:
            return False

        # if the buffer still has room, add the item
        if buffer_total_size < buffer_size:
            buffer[key_val].append(item)
            buffer_total_size += 1
        else:
            return False

        next(peekable_it)
        return True

    while _maybe_consume_from_iter():
        ...

    production_counter = 0

    def _next_key() -> T | None:
        nonlocal buffer_total_size, production_counter
        if not buffer_total_size:
            return None

        # TODO: add warning about buffer size not big enough for diversity_goal
        buffer_slot = production_counter % len(buffer)
        key = next(itertools.islice(buffer.keys(), buffer_slot, buffer_slot + 1))
        assert key in buffer
        key_buffer = buffer[key]

        buffer_total_size -= 1
        item = key_buffer.pop(0)
        if not key_buffer:
            del buffer[key]

        # try to fill the removed buffer slot
        _maybe_consume_from_iter()
        production_counter += 1
        return item

    while (item := _next_key()) is not None:
        yield item


def plan_read(
    uri: str,
    columns: list[str] | None = None,
    *,
    read_version: int | None = None,
    batch_size: int = 512,
    filter: str | None = None,  # noqa: A002
    shuffle_buffer_size: int = 0,
    task_shuffle_diversity: int | None = None,
    num_frags: int | None = None,
    **kwargs,
) -> Iterator[ReadTask]:
    """Make Plan for Reading Data from a Dataset"""
    it = _plan_read(
        uri,
        columns=columns,
        read_version=read_version,
        batch_size=batch_size,
        filter=filter,
        num_frags=num_frags,
    )
    # same as no shuffle
    if shuffle_buffer_size > 1 and task_shuffle_diversity is None:
        it = _buffered_shuffle(it, buffer_size=shuffle_buffer_size)
    elif task_shuffle_diversity is not None:
        buffer_size = max(4 * task_shuffle_diversity, shuffle_buffer_size)
        it = diversity_aware_shuffle(
            it,
            key=lambda task: task.checkpoint_key(),
            diversity_goal=task_shuffle_diversity,
            buffer_size=buffer_size,
        )

    return _LanceReadPlanIterator(
        it, _num_tasks(uri=uri, read_version=read_version, batch_size=batch_size)
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import hashlib
import logging
import os
import subprocess
import sys
from pathlib import Path
from typing import Literal

import attrs
import docker
import docker.models
import docker.models.images
import docker.utils
import docker.utils.json_stream
from ._state import PackagingState
from jinja2 import Environment, FileSystemLoader

_CUDA_BASE = "nvidia/cuda:12.4.1-cudnn-runtime-ubuntu22.04"
_CPU_BASE = "ubuntu:24.04"


_LOG = logging.getLogger(__name__)


def _get_base_dockerfile(
    backend: Literal["dataflow"],
    python_tool_chain: Literal["uv"],
    additional_pip_dependencies: list[str] | None = None,
    cuda: bool = False,
) -> str:
    if additional_pip_dependencies is None:
        additional_pip_dependencies = []
    module_location = Path(__file__).parent
    env = Environment(loader=FileSystemLoader(module_location), autoescape=True)
    # Load a template
    template = env.get_template(f"_base_{python_tool_chain}_{backend}.Dockerfile.j2")

    post_install_command = ""

    if additional_pip_dependencies:
        packages = " ".join(additional_pip_dependencies)
        post_install_command = f"RUN pip install {packages}"

    context = {
        "base_image": _CUDA_BASE if cuda else _CPU_BASE,
        "post_install_command": post_install_command,
        "python_version": f"{sys.version_info.major}.{sys.version_info.minor}",
    }
    return template.render(context)


def _get_incremental_dockerfile(
    backend: Literal["dataflow"],
    python_tool_chain: Literal["uv"],
    incremental_base: str,
) -> str:
    module_location = Path(__file__).parent
    env = Environment(loader=FileSystemLoader(module_location), autoescape=True)
    template = env.get_template(
        f"_incremental_{python_tool_chain}_{backend}.Dockerfile.j2"
    )

    context = {"incremental_base_image": incremental_base}

    return template.render(context)


@attrs.define
class DockerWorkspacePackager:
    """Local workspace docker packager"""

    # Docker client
    client: docker.DockerClient = attrs.field(
        factory=docker.from_env,
        kw_only=True,
    )

    additional_pip_dependencies: list[str] = attrs.field(factory=list, kw_only=True)

    # Path to the packaging cache
    state_store_path: Path = attrs.field(
        default=Path(os.environ.get("XDG_CONFIG_DIR", "~/.cache"))
        / "_geneva/ws_packager",
        converter=Path,
        kw_only=True,
    )

    _state: PackagingState = attrs.field(init=False)

    # support only uv for now
    python_tool_chain: Literal["uv"] = attrs.field(default="uv", kw_only=True)

    registry: str = attrs.field(
        default=os.environ.get("GENEVA_DOCKER_REGISTRY", ""), kw_only=True
    )

    image_name: str = attrs.field(
        default=os.environ.get("GENEVA_DOCKER_IMAGE_NAME", ""), kw_only=True
    )

    def __attrs_post_init__(self) -> None:
        self._state = PackagingState(self.state_store_path)

    def _image_name(self, tag: str) -> str:
        return f"{self.registry}/{self.image_name}:{tag}"

    def build(
        self,
        image_tag: str,
        cuda: bool = False,
        platform: Literal["linux/amd64", "linux/arm64"] = "linux/amd64",
    ) -> str:
        """Build a docker image"""

        # this context determines which base image (if there is one)
        # to use for incremental builds
        #
        # sort the item tuples so that the hash is deterministic
        # from potential order changes
        built_context = sorted(
            {
                "platform": platform,
                "cuda": cuda,
                "backend": "dataflow",
                "additional_pip_dependencies": self.additional_pip_dependencies,
                "python_tool_chain": self.python_tool_chain,
                "python_version": sys.version_info,
            }.items()
        )

        context_hash = hashlib.md5(str(built_context).encode()).hexdigest()

        if (
            incremental_image := self._state.current_incremental_image(context_hash)
        ) is not None:
            dockerfile = _get_incremental_dockerfile(
                backend="dataflow",
                python_tool_chain=self.python_tool_chain,
                incremental_base=incremental_image,
            )
            is_incremental = True
        else:
            dockerfile = _get_base_dockerfile(
                backend="dataflow",
                python_tool_chain=self.python_tool_chain,
                additional_pip_dependencies=self.additional_pip_dependencies,
                cuda=cuda,
            )
            is_incremental = False

        image_name = self._image_name(image_tag)
        subprocess.run(
            [
                "docker",
                "build",
                "-f",
                "-",
                "-t",
                image_name,
                "--platform",
                platform,
                ".",
            ],
            input=dockerfile.encode(),
            env={"DOCKER_BUILDKIT": "1", **os.environ.copy()},
            check=True,
        )

        # avoid setting incremental image on incremental builds
        # this is because the last layer of the incremental image
        # could be very large and overlaps with the incoming changes
        # TODO: we should trigger a full build of the incremental
        # base image based on some heuristic, e.g. change size, deps, etc.
        if not is_incremental:
            self._state.set_incremental_image(context_hash, image_name)

        return image_name

    def push(self, image_tag: str) -> None:
        """Push a docker image"""
        self.client.images.push(self._image_name(image_tag))


def register_packager_parser(subparsers) -> None:
    packager_parser = subparsers.add_parser(
        "package", description="Build local workspace and push docker image"
    )
    packager_parser.add_argument(
        "--docker-image",
        dest="image_name",
        required=True,
        help="The name of the image to build.",
    )
    packager_parser.add_argument(
        "--cuda",
        dest="cuda",
        action="store_true",
        default=False,
        help="Use CUDA base image.",
    )
    packager_parser.add_argument(
        "--platform",
        dest="platform",
        choices=["linux/amd64", "linux/arm64"],
        help="The platform to build the image for.",
    )
    packager_parser.set_defaults(func=run)


def run(args) -> None:
    packager = DockerWorkspacePackager()
    image = packager.build(
        image_tag=args.image_name,
        cuda=args.cuda,
        platform=args.platform,
    )

    _LOG.info("Built image %s, pushing to %s", image.id, args.image_name)
    packager.client.images.push(args.image_name)
    _LOG.info("Pushed image %s", args.image_name)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import itertools
from pathlib import Path

import lance
import pyarrow as pa
import pytest

from geneva import LanceCheckpointStore, udf
from geneva.runners.ray.pipeline import run_ray_add_column


@pytest.mark.parametrize(
    "shuffle_config",
    [
        {
            "applier_batch_size": applier_batch_size,
            "batch_size": batch_size,
            "shuffle_buffer_size": shuffle_buffer_size,
            "task_shuffle_diversity": task_shuffle_diversity,
        }
        for (
            applier_batch_size,
            batch_size,
            shuffle_buffer_size,
            task_shuffle_diversity,
        ) in itertools.product(
            [1, 2, 3],
            range(1, 5),
            [0, 1, 3, 16],
            [None, 1, 3, 16],
        )
    ],
)
def test_run_ray_add_column(tmp_path: Path, shuffle_config) -> None:
    @udf(data_type=pa.int32(), batch_size=8)
    def add_one(a) -> int:
        return a + 1

    data = {"a": pa.array(range(256))}
    tbl = pa.Table.from_pydict(data)
    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    tbl_path = tmp_path / "foo.lance"
    ds = lance.write_dataset(tbl, tbl_path, max_rows_per_file=16)

    def return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(
            {"b": pa.array([None] * batch.num_rows, pa.int32())}
        )

    new_frags = []
    new_schema = None
    for frag in ds.get_fragments():
        new_fragment, new_schema = frag.merge_columns(return_none, columns=["a"])
        new_frags.append(new_fragment)

    assert new_schema is not None
    merge = lance.LanceOperation.Merge(new_frags, new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)

    run_ray_add_column(
        str(tbl_path), ["a"], {"b": add_one}, ckp_store, **shuffle_config
    )

    ds = lance.dataset(tbl_path)
    assert ds.to_table().to_pydict() == {
        "a": list(range(256)),
        "b": [x + 1 for x in range(256)],
    }



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from pathlib import Path

import pyarrow as pa

from geneva.query import Scan, connect


def test_scan_over_fragments(tmp_path: Path) -> None:
    db = connect(tmp_path)

    a = pa.array([1, 2, 3])
    b = pa.array([4, 5, 6])
    tbl = db.create_table("tbl", pa.Table.from_arrays([a, b], names=["a", "b"]))

    c = pa.array([7, 8, 9])
    d = pa.array([10, 11, 12])
    tbl.add(pa.Table.from_arrays([c, d], names=["a", "b"]))

    fragments = tbl.get_fragments()
    assert len(fragments) == 2

    scan = Scan(tbl, columns=["a"]).with_fragments(fragments[0].fragment_id)
    results = list(scan.to_batches())

    assert len(results) == 1
    assert results[0]["a"].equals(a)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# module for configuration
import abc
import inspect
import itertools
import logging
import os
from pathlib import Path
from types import UnionType

from typing_extensions import Self

from geneva.config.loader import (
    ConfigLoader,
    ConfigResolver,
    chain,
    from_env,
    from_file,
    from_pyproject,
    loader,
)

_LOG = logging.getLogger(__name__)


class ConfigBase(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """Return the name of the configuration dict to extract"""

    @classmethod
    def from_loader(cls, data: ConfigLoader) -> Self:
        """Populate the configuration from a loader"""
        subloader = data.get(cls.name(), {})

        args = {}
        for arg, arg_type in inspect.get_annotations(cls).items():
            # handle Optional
            is_optional = False
            if isinstance(arg_type, UnionType):
                if len(arg_type.__args__) == 2 and arg_type.__args__[1] is type(None):
                    arg_type = arg_type.__args__[0]
                    is_optional = True
                else:
                    raise ValueError(f"Union type {arg_type} not supported")
            if issubclass(arg_type, ConfigBase):
                try:
                    args[arg] = arg_type.from_loader(subloader)
                except KeyError:
                    if not is_optional:
                        raise
                    _LOG.debug(
                        f"Optional key {arg} not found in {cls.name()},"
                        " treating as None"
                    )
            else:
                # if the key is not present, the default value is used, don't pass None
                if (value := subloader.get(arg, None)) is not None:
                    args[arg] = value

        return cls(**args)


_CONFIG_DIR = Path(os.environ.get("GENEVA_CONFIG_DIR", "./.config")).absolute()

_CONFIG_CHAIN = chain(
    from_env(),
    from_pyproject(),
    *[
        from_file(Path(f))
        for f in itertools.chain(
            _CONFIG_DIR.glob("*.json"),
            _CONFIG_DIR.glob("*.yaml"),
            _CONFIG_DIR.glob("*.yml"),
            _CONFIG_DIR.glob("*.toml"),
        )
    ],
)

CONFIG_LOADER = loader(_CONFIG_CHAIN)


def override_config(config: ConfigResolver) -> None:
    """Add a configuration override, which will be applied first"""
    global _CONFIG_CHAIN
    _CONFIG_CHAIN.push_front(config)


def default_config(config: ConfigResolver) -> None:
    """Add a configuration defaults, which will be applied last"""
    global _CONFIG_CHAIN
    _CONFIG_CHAIN.push_back(config)


__all__ = ["ConfigBase", "CONFIG_LOADER", "override_config", "default_config"]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import itertools
from pathlib import Path
from typing import Any

import lance
import pyarrow as pa
import pytest

from geneva import LanceCheckpointStore, udf
from geneva.runners.dataflow.dataflow import DataflowOptions, run_dataflow_add_column


@udf(data_type=pa.int32(), batch_size=8)
def add_one(a) -> int:
    return a + 1


@pytest.mark.parametrize(
    "shuffle_config",
    [
        {
            "batch_size": batch_size,
            "shuffle_buffer_size": shuffle_buffer_size,
            "task_shuffle_diversity": task_shuffle_diversity,
        }
        for (
            batch_size,
            shuffle_buffer_size,
            task_shuffle_diversity,
        ) in itertools.product(
            range(1, 5),
            [0, 1, 3, 16],
            [None, 1, 3, 16],
        )
    ],
)
def test_run_dataflow_add_column(
    tmp_path: Path, shuffle_config: dict[str, Any]
) -> None:
    data = {"a": pa.array(range(1024))}
    tbl = pa.Table.from_pydict(data)
    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    tbl_path = tmp_path / "foo.lance"
    ds = lance.write_dataset(tbl, tbl_path, max_rows_per_file=16)

    def return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(
            {"b": pa.array([None] * batch.num_rows, pa.int32())}
        )

    new_frags = []
    new_schema = None
    for frag in ds.get_fragments():
        new_fragment, new_schema = frag.merge_columns(return_none, columns=["a"])
        new_frags.append(new_fragment)

    assert new_schema is not None
    merge = lance.LanceOperation.Merge(new_frags, new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)

    run_dataflow_add_column(
        str(tbl_path),
        ["a"],
        {"b": add_one},
        ckp_store,
        DataflowOptions(
            runner="DirectRunner",
        ),
        test_run=False,
        **shuffle_config,
    ).result.wait_until_finish()

    ds = lance.dataset(tbl_path)

    assert ds.to_table().to_pydict() == {
        "a": list(range(1024)),
        "b": [x + 1 for x in range(1024)],
    }


def test_dataflow_runner_test_run(tmp_path: Path) -> None:
    data = {"a": pa.array(range(100))}
    tbl = pa.Table.from_pydict(data)
    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    tbl_path = tmp_path / "foo.lance"
    # do it twice to have two fragments
    ds = lance.write_dataset(tbl, tbl_path)

    data = {"a": pa.array(range(100, 200))}
    tbl = pa.Table.from_pydict(data)
    ds = lance.write_dataset(tbl, tbl_path, mode="append")

    def return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(
            {"b": pa.array([None] * batch.num_rows, pa.int32())}
        )

    new_frags = []
    new_schema = None
    for frag in ds.get_fragments():
        new_fragment, new_schema = frag.merge_columns(return_none, columns=["a"])
        new_frags.append(new_fragment)

    assert new_schema is not None
    merge = lance.LanceOperation.Merge(new_frags, new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)

    run_dataflow_add_column(
        str(tbl_path),
        ["a"],
        {"b": add_one},
        ckp_store,
        DataflowOptions(
            runner="DirectRunner",
        ),
        batch_size=8,
        test_run=True,
    ).result.wait_until_finish()

    ds = lance.dataset(tbl_path)

    assert ds.to_table().to_pydict() == {
        "a": list(range(200)),
        "b": [x + 1 for x in range(100)] + [None] * 100,
    }



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# lance dataset distributed transform job checkpointing + UDF utils

from geneva.apply import LanceRecordBatchUDFApplier, ReadTask
from geneva.checkpoint import (
    ArrowFsCheckpointStore,
    CheckpointStore,
    InMemoryCheckpointStore,
    LanceCheckpointStore,
)
from geneva.db import connect
from geneva.docker import DockerWorkspacePackager
from geneva.transformer import udf

__all__ = [
    "ArrowFsCheckpointStore",
    "CheckpointStore",
    "connect",
    "InMemoryCheckpointStore",
    "LanceRecordBatchUDFApplier",
    "LanceCheckpointStore",
    "ReadTask",
    "udf",
    "DockerWorkspacePackager",
]

__version__ = "0.1.7"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# reshuffle tasks by fragment ID and batch them together

import apache_beam as beam
from apache_beam import PTransform, pvalue, typehints
from apache_beam.runners.pipeline_context import PipelineContext

from geneva.apply import ReadTask

_MAX_BUCKET_SIZE = 8192


@typehints.with_input_types(ReadTask)
@typehints.with_output_types(ReadTask)
class ReshuffleByFragmentChunks(PTransform):
    """
    Reshuffle tasks by fragment ID and batch them together.
    """

    _URN = "LanceShuffleByFragmentChunks"

    def __init__(self, bucket_size=_MAX_BUCKET_SIZE) -> None:
        self.bucket_size = bucket_size

    def expand(
        self,
        input_or_inputs: pvalue.PValue,
    ) -> pvalue.PCollection:
        return (
            input_or_inputs
            | "AddFragIDandChunkIdKey"
            >> beam.Map(lambda x: ((x.frag_id, x.offset // self.bucket_size), x))
            | "GroupByFragIDandChunkId" >> beam.GroupByKey()
            | "ExtractTask" >> beam.FlatMap(lambda x: x[1])
        )  # type: ignore

    def to_runner_api_parameter(
        self,
        unused_context: PipelineContext,
    ) -> tuple[str, None]:
        return self._URN, None

    @staticmethod
    @PTransform.register_urn(_URN, None)
    def from_runner_api_parameter(
        unused_ptransform,
        unused_parameter,
        unused_context,
    ) -> "ReshuffleByFragmentChunks":
        return ReshuffleByFragmentChunks()



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import attrs


@attrs.define
class RayOptions:
    docker_image: str | None = attrs.field(
        default=None,
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# module for packaging workspace and UDFs
# also data spec for persisting the artifacts

import abc
import base64
import hashlib
import json
import logging
import uuid

import attrs
import cloudpickle
from typing_extensions import Self

from geneva import DockerWorkspacePackager
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


class UDFBackend(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def packager(cls) -> "UDFPackager":
        """Return the packager for this backend."""

    def to_bytes(self) -> bytes:
        return json.dumps(attrs.asdict(self)).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        return cls(**json.loads(data.decode()))


@attrs.define
class UDFSpec:
    """Specification for a user-defined function.

    This is an holder of an arbitrary user-defined function,
    which can use an backend for marshalling.

    The most common is likely Docker + some kind of workspace
    persistence. However, we want to support more than just
    Docker, so we create this "out most" abstraction to allow
    for more flexibility.
    """

    # the name of the udf
    name: str = attrs.field(validator=attrs.validators.min_len(1))

    # the packaging backend for the udf
    backend: str = attrs.field()

    @backend.validator
    def _validate_backend(self, attribute, value) -> None:
        backend_names = [cls.__name__ for cls in UDFBackend.__subclasses__()]
        unique_backend_names = set(backend_names)

        # ruff: noqa: ERA001
        # TODO - deduplication not working right sometimes
        # if len(backend_names) != len(unique_backend_names):
        #     duplicates = []
        #     while unique_backend_names:
        #         name = backend_names.pop()
        #         if name not in unique_backend_names:
        #             duplicates.append(name)
        #         else:
        #             unique_backend_names.discard(name)
        #     raise ValueError(
        #         "Duplicate backend names found, "
        #         "please don't define the same backend name twice."
        #         f"\nDuplicates: {duplicates}"
        #     )

        if value not in unique_backend_names:
            raise ValueError(f"Unknown backend: {value}")

    udf_payload: bytes = attrs.field()

    # the payload for the runner -- This is a HACK for allowing phalanx knowing
    # how to dispatch the UDF job. Make sure changes here are compatible to
    # parsing in phalanx.
    runner_payload: bytes | None = attrs.field(default=None)

    @classmethod
    def udf_from_spec(cls, data) -> UDF:
        # TODO: load the spec and find the backend,
        # then call the packager to do the next level unmarshalling
        pass


@attrs.define
class DockerUDFSpecV1(UDFBackend):
    """Specification for a user-defined function that runs in a Docker container.
    -- Version 1

    In this packaging spec, the python interpreter is assumed to be correctly
    setup in the container, and the user-defined function is expected to load
    using cloudpickle. With the option of downloading additional workspace
    files from a remote location (S3, GCS, etc).
    """

    # the image to run the udf in
    image: str = attrs.field()

    # the tag of the image
    tag: str | None = attrs.field()

    # spec from https://docs.docker.com/reference/cli/docker/image/tag/#description
    @tag.validator
    def _validate_tag(self, attribute, value: str | None) -> None:
        if value is None:
            return
        if len(value) > 128:
            raise ValueError("Tag must be less than 128 characters.")
        if not all(c.isalpha() or c.isnumeric or c in {"_", ".", "-"} for c in value):
            raise ValueError("Tag must be valid alphanumeric.")

    # optionally have a zip of the workspace and store it separately
    # this should be the path to the zip file on (S3, GCS, etc)
    workspace_zip: str | None = attrs.field()

    # the checksum of the workspace zip
    workspace_checksum: str | None = attrs.field()

    @workspace_checksum.validator
    def _validate_workspace_checksum(self, attribute, value) -> None:
        if self.workspace_zip and not value:
            raise ValueError(
                "Workspace checksum must not be empty when a workspace is provided."
            )

    # the udf pickle to run
    udf_pickle: bytes = attrs.field()

    @udf_pickle.validator
    def _validate_udf_pickle(self, attribute, value) -> None:
        if not value:
            raise ValueError("UDF pickle must not be empty.")
        udf = cloudpickle.loads(value)
        if not isinstance(udf, UDF):
            raise ValueError("UDF pickle must contain a UDF object.")

    # the checksum of the udf pickle
    udf_checksum: str = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        self.udf_checksum = hashlib.sha256(self.udf_pickle).hexdigest()

    @classmethod
    def packager(cls) -> "UDFPackager":
        return DockerUDFPackager()

    def to_bytes(self) -> bytes:
        self_as_dict = attrs.asdict(self)
        self_as_dict["udf_pickle"] = base64.b64encode(
            self_as_dict["udf_pickle"]
        ).decode("utf-8")
        return json.dumps(self_as_dict).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        self_as_dict = json.loads(data.decode())
        self_as_dict["udf_pickle"] = base64.b64decode(
            self_as_dict["udf_pickle"].encode("utf-8")
        )

        checksum = self_as_dict.pop("udf_checksum")  # not part of the init
        val = cls(**self_as_dict)
        val.udf_checksum = checksum

        return val


class UDFPackager(abc.ABC):
    """Packager for user-defined functions."""

    @abc.abstractmethod
    def marshal(self, udf: UDF) -> UDFSpec:
        """Marshal a user-defined function."""

    @abc.abstractmethod
    def unmarshal(self, spec: UDFSpec) -> UDF:
        """Unmarshal a user-defined function."""


class DockerUDFPackager(UDFPackager):
    def marshal(self, udf: UDF) -> UDFSpec:
        image_tag = uuid.uuid4().hex

        packager = DockerWorkspacePackager()
        _LOG.info("Building UDF image: %s", image_tag)
        image_name = packager.build(image_tag, platform="linux/amd64", cuda=udf.cuda)
        tag = image_name.split(":")[-1]
        image_name = ":".join(image_name.split(":")[:-1])
        _LOG.info("Pushing UDF image: %s", image_tag)
        packager.push(image_tag)
        _LOG.info("Pushed UDF image: %s", image_tag)

        udf_pickle = cloudpickle.dumps(udf)

        return UDFSpec(
            name=udf.name,
            backend=DockerUDFSpecV1.__name__,
            udf_payload=DockerUDFSpecV1(
                image=image_name,
                tag=tag,
                workspace_zip=None,
                workspace_checksum=None,
                udf_pickle=udf_pickle,
            ).to_bytes(),
            runner_payload=json.dumps(
                {
                    "image": image_name,
                }
            ).encode(),
        )

    def unmarshal(self, spec: UDFSpec) -> UDF:
        if spec.backend != DockerUDFSpecV1.__name__:
            raise ValueError("Invalid backend for UDF spec.")

        docker_spec = DockerUDFSpecV1.from_bytes(spec.udf_payload)
        udf = cloudpickle.loads(docker_spec.udf_pickle)
        if not isinstance(udf, UDF):
            raise ValueError("UDF pickle must contain a UDF object.")
        return udf



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# Use lance lake with Google Dataflow

import logging
import os
import re
import uuid
from collections.abc import Callable, Iterator

import more_itertools

from geneva import DockerWorkspacePackager

try:
    import apache_beam as beam
    from apache_beam.internal import pickler
    from apache_beam.options.pipeline_options import (
        GoogleCloudOptions,
        PipelineOptions,
        SetupOptions,
        StandardOptions,
    )
    from apache_beam.pipeline import Pipeline
    from apache_beam.transforms.combiners import Top
except ImportError as err:
    raise ImportError(
        "apache_beam is required for this module, pip install geneva[dataflow]",
    ) from err


import attrs
import lance
import lance.commit
import pyarrow as pa

from geneva.apply import LanceRecordBatchUDFApplier, plan_read
from geneva.checkpoint import CheckpointStore
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.runners.dataflow.reshuffle import ReshuffleByFragmentChunks
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


@attrs.define
class DataflowOptions:
    """Dataflow Options"""

    runner: str = attrs.field(
        default="DataflowRunner",
    )
    region: str = attrs.field(
        default="us-central1",
    )
    project: str = attrs.field(
        default=os.environ.get("GENEVA_DATAFLOW_PROJECT", ""),
    )
    temp_location: str = attrs.field(
        default=os.environ.get(
            "GENEVA_DATAFLOW_TEMP_LOCATION",
            f"{os.environ.get('GENEVA_DATAFLOW_GCS', '')}/temp",
        ),
    )
    staging_location: str = attrs.field(
        default=os.environ.get(
            "GENEVA_DATAFLOW_STAGING_LOCATION",
            f"{os.environ.get('GENEVA_DATAFLOW_GCS', '')}/staging",
        ),
    )
    docker_image: str | None = attrs.field(
        default=os.environ.get("GENEVA_DATAFLOW_DOCKER_IMAGE"),
    )
    disk_size_gb: int = attrs.field(
        default=100,
    )
    machine_type: str = attrs.field(
        default=os.environ.get("GENEVA_DATAFLOW_MACHINE_TYPE", "n2-highmem-16"),
    )

    _GPU_MACHINES = [
        "g2",  # L4
        # "a2", disallow A100 for now
        # "a3", disallow H100 for now
    ]

    _ARM_MACHINES = [
        "c4a",
    ]

    _X86_MACHINES = [
        "n2",
        "c3",
        "c3d",
        "c4",
    ]

    _ALL_MACHINES = _GPU_MACHINES + _ARM_MACHINES + _X86_MACHINES

    @property
    def is_arm(self) -> bool:
        return any(
            self.machine_type.startswith(f"{machine}-")
            for machine in self._ARM_MACHINES
        )

    @machine_type.validator
    def _validate_machine_type(self, attribute, value) -> None:
        if not any(value.startswith(f"{machine}-") for machine in self._ALL_MACHINES):
            raise ValueError(
                f"machine_type must start with one of {self._ALL_MACHINES}"
            )

    @property
    def is_cuda(self) -> bool:
        return any(
            self.machine_type.startswith(f"{machine}-")
            for machine in self._GPU_MACHINES
        )

    @property
    def additional_options(self) -> list[str]:
        if self.is_cuda:
            return [
                # enable GPU
                "--dataflow_service_options=worker_accelerator=type:nvidia-l4;count:1;install-nvidia-driver;use_nvidia_mps",
                # make sure we don't spin up tons of containers and waste memory
                "--experiments=no_use_multiple_sdk_containers",
            ]
        if self.runner in {"DirectRunner", "direct"}:
            return [
                "--direct_num_workers=0",
                "--direct_running_mode=multi_threading",
            ]
        return []


def map_with_value(  # noqa: ANN201
    fn: Callable,
):
    def _wrapped(x):  # noqa: ANN202
        return x, fn(x)

    return _wrapped


def extract_frag_id_as_key(x):  # noqa: ANN201
    return x[0].frag_id, x


def write_fragment(  # noqa: ANN201
    uri: str,
    column_name: str,
    store: CheckpointStore,
):
    dataset = lance.dataset(uri)
    # MASSIVE HACK: open up an API to get the field id from the column name
    field_id = re.compile(rf'name: "{column_name}", id: (?P<field_id>[\d]*),').findall(
        str(dataset.lance_schema)
    )
    assert len(field_id) == 1
    field_id = int(field_id[0])

    def _write_fragment(x):  # noqa: ANN202
        frag_id, checkpoint_keys = x
        _LOG.info("Writing fragment %s, with %d keys", frag_id, len(checkpoint_keys))
        batches: list[dict[str, str]] = list(checkpoint_keys)

        def _iter() -> Iterator[pa.RecordBatch]:
            for batch in batches:
                res = {key: store[value]["data"] for key, value in batch.items()}
                yield pa.RecordBatch.from_pydict(res)

        it = more_itertools.peekable(_iter())
        rbr = pa.RecordBatchReader.from_batches(it.peek().schema, it)

        # TODO: this doesn't support struct or complex schema yet
        new_data = lance.fragment.write_fragments(
            rbr,
            uri,
            max_rows_per_file=1 << 31,
            max_bytes_per_file=1 << 40,
        )

        assert len(new_data) == 1
        new_data = new_data[0]
        assert len(new_data.files) == 1
        new_datafile = new_data.files[0]

        new_datafile.fields = [field_id]
        new_datafile.column_indices = [0]

        return frag_id, new_datafile

    return _write_fragment


def sort_and_collect_batch(x):  # noqa: ANN201
    (frag_id, data) = x
    _LOG.info("Sorting %d keys for frag: %s", len(data), frag_id)
    data = list(data)
    data.sort(key=lambda x: x[0].offset)

    return frag_id, [x[1] for x in data]


def commit_fragments(  # noqa: ANN201
    uri: str,
    version: int,
):
    def _commit_fragments(x) -> None:
        _LOG.info("Committing %d fragments", len(x))
        operation = lance.LanceOperation.DataReplacement(
            replacements=[
                lance.LanceOperation.DataReplacementGroup(
                    fragment_id=fragment_id,
                    new_file=new_file,
                )
                for (fragment_id, new_file) in x
            ]
        )
        lance.LanceDataset.commit(uri, operation, read_version=version)

    return _commit_fragments


def run_dataflow_add_column(
    uri: str,
    columns: list[str],
    transforms: dict[str, UDF],
    checkpoint_store: CheckpointStore,
    dataflow_options: DataflowOptions,
    /,
    job_id: str | None = None,
    batch_size: int = 512,
    read_version: int | None = None,
    test_run: bool = True,
    **kwargs,
) -> Pipeline | None:
    is_cuda = any(udf.cuda for udf in transforms.values())

    if is_cuda and not dataflow_options.is_cuda:
        _LOG.info(
            "CUDA UDFs detected, but dataflow machine type is not cuda,"
            " setting machine type to use L4 GPU"
        )
        dataflow_options.machine_type = "g2-standard-8"

    if (
        dataflow_options.docker_image is None
        and dataflow_options.runner == "DataflowRunner"
    ):
        _LOG.warning("No docker image specified, building one")
        tag = uuid.uuid4().hex
        packager = DockerWorkspacePackager(**kwargs)
        docker_image = packager.build(tag, platform="linux/amd64", cuda=is_cuda)
        packager.push(tag)
    else:
        docker_image = dataflow_options.docker_image

    pipeline_args = [
        "--runner",
        dataflow_options.runner,
        "--project",
        dataflow_options.project,
        "--region",
        dataflow_options.region,
        "--temp_location",
        dataflow_options.temp_location,
        "--staging_location",
        dataflow_options.staging_location,
        "--experiments=use_runner_v2",
        "--dataflow_service_options=enable_dynamic_thread_scaling",
        # we use flexrs, meaning we enable dataflow shuffle service
        # we don't need a lot of disk space
        "--disk_size_gb=75",
        # TODO: DO NOT HARDCODE HARDWARE CONFIG
        "--machine_type=" + dataflow_options.machine_type,
        "--flexrs_goal=SPEED_OPTIMIZED",
        "--max_num_workers=1024",
        "--dataflow_service_options=min_num_workers=4",
        *dataflow_options.additional_options,
    ]
    if dataflow_options.runner == "DataflowRunner":
        if docker_image is None:
            raise ValueError("Docker image must be specified for DataflowRunner")
        pipeline_args += ["--sdk_container_image=" + docker_image]

    try:
        import prefect
        from prefect.exceptions import MissingContextError

        try:
            job_id = job_id or prefect.context.get_run_context().flow_run.id.hex
            _LOG.info("Using prefect job id: %s", job_id)
        except MissingContextError:
            job_id = None
    except ImportError:
        job_id = None

    job_id = job_id or uuid.uuid4().hex

    # We use the save_main_session option because one or more DoFn's in this
    # workflow rely on global context (e.g., a module imported at module level).
    pipeline_options = PipelineOptions(pipeline_args)
    # we usually submit from a local environment, don't want save the main session
    pipeline_options.view_as(SetupOptions).save_main_session = False
    # always use cloudpickle
    pipeline_options.view_as(SetupOptions).pickle_library = pickler.USE_CLOUDPICKLE
    # otherwise __exit__ will wait for the job to finish
    pipeline_options.view_as(StandardOptions).no_wait_until_finish = True
    # set the job name

    applier = LanceRecordBatchUDFApplier(
        udfs=transforms,
        checkpoint_store=checkpoint_store,
        error_logger=CheckpointStoreErrorLogger(job_id, checkpoint_store),
    )

    if read_version is None:
        read_version = lance.dataset(uri).version

    output_column = list(transforms.keys())[0]

    table_name = uri.split("/")[-1].split(".")[0]
    name = f"{table_name}-{output_column}-{job_id[:8]}"
    pipeline_options.view_as(GoogleCloudOptions).job_name = name.replace("_", "-")
    # The pipeline will be run on exiting the with block.
    with Pipeline(options=pipeline_options) as p:
        pcol = (
            p
            | beam.Create([None])
            | "CreateTasks"
            >> beam.FlatMap(
                lambda _: plan_read(uri, columns, batch_size=batch_size, **kwargs)
            )
        )

        if test_run:
            pcol = (
                pcol
                | "ExtractFragIDForSample" >> beam.Map(lambda x: (x.frag_id, x))
                | "GroupByFragIDForSample" >> beam.GroupByKey()
                | "TakeSmallestFragID" >> Top.Smallest(1)
                | beam.FlatMap(lambda x: x[0][1])
            )

        pcol = (
            pcol
            | ReshuffleByFragmentChunks()
            | "Apply UDFs" >> beam.Map(map_with_value(applier.run))
            | "ExtractFragID" >> beam.Map(extract_frag_id_as_key)
            | "GroupByKey" >> beam.GroupByKey()
            | "SortAndCollectBatch" >> beam.Map(sort_and_collect_batch)
            | "WriteFragments"
            >> beam.Map(write_fragment(uri, output_column, checkpoint_store))
            | "Collect" >> beam.combiners.ToList()
            | "Commit" >> beam.Map(commit_fragments(uri, read_version))
        )

    return p



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from pathlib import Path

import cloudpickle
import lance
import pyarrow as pa
import pyarrow.compute as pc
import pytest

from geneva import InMemoryCheckpointStore, connect, udf
from geneva.apply import LanceRecordBatchUDFApplier, plan_read
from geneva.debug.logger import CheckpointStoreErrorLogger


def test_create_plan(tmp_path: Path) -> None:
    db = connect(tmp_path)
    tbl = db.create_table("tbl", pa.table({"a": [1, 2, 3]}))

    plans = list(plan_read(tbl.uri, batch_size=16))
    assert len(plans) == 1
    plan = plans[0]
    assert plan.uri == tbl.uri
    assert plan.offset == 0
    assert plan.limit == 3


def test_create_plan_with_diverse_shuffle(tmp_path: Path) -> None:
    ds = lance.write_dataset(
        pa.table({"a": range(1024)}), tmp_path / "tbl", max_rows_per_file=16
    )

    plans = list(plan_read(ds.uri, batch_size=1, task_shuffle_diversity=4))
    assert len(plans) == 1024
    plan = plans[0]
    assert plan.uri == ds.uri
    assert plan.offset == 0
    assert plan.limit == 1


@udf(input_columns=["a"])
def one(*args, **kwargs) -> int:
    return 1


def test_applier(tmp_path: Path) -> None:
    db = connect(tmp_path)
    tbl = db.create_table("tbl", pa.table({"a": [1, 2, 3]}))

    plans = list(plan_read(tbl.uri, batch_size=16))
    assert len(plans) == 1
    plan = plans[0]
    assert plan.uri == tbl.uri
    assert plan.offset == 0
    assert plan.limit == 3

    store = InMemoryCheckpointStore()
    applier = LanceRecordBatchUDFApplier(
        udfs={"one": one},
        checkpoint_store=store,
    )
    result = applier.run(plan)
    assert len(result) == 1
    batch = pa.RecordBatch.from_pydict(
        {name: store[key]["data"] for name, key in result.items()}
    )
    assert len(batch) == 3
    assert batch.to_pydict() == {"one": [1, 1, 1]}


@udf()
def errors_on_three(a: int) -> int:
    if a == 3:
        raise ValueError("This is an error")
    return 1


def test_applier_error_logging(tmp_path: Path) -> None:
    db = connect(tmp_path)
    tbl = db.create_table("tbl", pa.table({"a": [1, 2, 3]}))

    plans = list(plan_read(tbl.uri, batch_size=16))
    assert len(plans) == 1
    plan = plans[0]
    assert plan.uri == tbl.uri
    assert plan.offset == 0
    assert plan.limit == 3

    store = InMemoryCheckpointStore()
    error_logger = CheckpointStoreErrorLogger("job_id", store)
    applier = LanceRecordBatchUDFApplier(
        udfs={
            "one": errors_on_three,
        },
        checkpoint_store=store,
        error_logger=error_logger,
    )
    with pytest.raises(RuntimeError):
        applier.run(plan)

    assert len(list(error_logger.list_errors())) == 1
    error_id = list(error_logger.list_errors())[0]
    error = error_logger.get_error_row(error_id).to_pylist()[0]
    assert error["error"] == "This is an error"
    assert error["udf"] == cloudpickle.dumps(errors_on_three)
    assert error["seq"] == 0


def test_plan_with_filter(tmp_path: Path) -> None:
    db = connect(tmp_path)

    tbl = db.create_table("t", pa.table({"a": range(100)}))
    tbl.add(pa.table({"a": range(100, 200)}))
    tbl.add(pa.table({"a": range(200, 300)}))
    tbl.add(pa.table({"a": range(300, 400)}))

    fragments = tbl.get_fragments()
    assert len(fragments) == 4

    tasks = list(plan_read(tbl.uri, filter="a > 100 AND a % 2 == 0"))
    assert len(tasks) == 3


def test_udf_with_arrow_params(tmp_path: Path) -> None:
    @udf(data_type=pa.int32())
    def batch_udf(a: pa.Array, b: pa.Array) -> pa.Array:
        assert a == pa.array([1, 2, 3])
        assert b == pa.array([4, 5, 6])
        return pc.add(a, b)

    db = connect(tmp_path)
    tbl = db.create_table("t", pa.table({"a": [1, 2, 3], "b": [4, 5, 6]}))

    store = InMemoryCheckpointStore()
    applier = LanceRecordBatchUDFApplier(
        udfs={
            "c": batch_udf,
        },
        checkpoint_store=store,
    )
    result = applier.run(next(plan_read(tbl.uri, batch_size=16)))
    assert len(result) == 1
    batch = pa.RecordBatch.from_pydict(
        {name: store[key]["data"] for name, key in result.items()}
    )
    assert batch == pa.RecordBatch.from_pydict(
        {"c": pa.array([5, 7, 9], type=pa.int32())},
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# Local workspace docker packaging module
from geneva.docker.packager import DockerWorkspacePackager

__all__ = ["DockerWorkspacePackager"]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# Module for tracking local packaging cache state
# Private to the docker packaging module
import os
from pathlib import Path

import attrs

from geneva.utils.sqlitekv import SQLiteKV


@attrs.define
class PackagingState:
    """Local packaging cache state"""

    # Path to the packaging cache
    store_path: Path = attrs.field(
        default=Path(os.environ.get("XDG_CONFIG_DIR", "~/.cache"))
        / "_geneva/ws_packager",
        converter=Path,
    )

    _SQLITE_NAME = "state.db"
    _state_store: SQLiteKV = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        self._state_store = SQLiteKV(
            self.store_path.expanduser().resolve() / self._SQLITE_NAME,
        )

    def current_incremental_image(self, context_hash: str) -> str | None:
        return self._state_store.get(f"{context_hash}#last_pushed_image", None)

    def set_incremental_image(self, context_hash: str, image_name: str) -> None:
        self._state_store[f"{context_hash}#last_pushed_image"] = image_name



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import heapq
import logging
import re
import uuid
from collections.abc import Iterator
from typing import cast

import attrs
import lance
import lance.file
import pyarrow as pa
import ray.actor
import ray.data
import ray.util.queue
from tqdm.auto import tqdm

from geneva.apply import LanceRecordBatchUDFApplier, ReadTask, plan_read
from geneva.checkpoint import CheckpointStore
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


def _combine_chunks(arr: pa.Array) -> pa.Array:
    if isinstance(arr, pa.ChunkedArray):
        assert len(arr.chunks) == 1
        return arr.chunks[0]
    return arr


@ray.remote(num_cpus=1)
@attrs.define
class ApplierActor:
    applier: LanceRecordBatchUDFApplier

    def run(self, task) -> tuple[ReadTask, dict[str, str]]:
        return task, self.applier.run(task)


ApplierActor: ray.actor.ActorClass = cast(ray.actor.ActorClass, ApplierActor)


@ray.remote
@attrs.define
class FragmentWriter:
    uri: str
    column_name: str
    store: CheckpointStore

    fragment_id: int

    checkpoint_keys: ray.util.queue.Queue

    def write(self) -> tuple[int, lance.fragment.DataFile]:
        dataset = lance.dataset(self.uri)
        import more_itertools

        buffer: list[tuple[int, dict[str, str]]] = []
        num_rows = dataset.get_fragment(self.fragment_id).count_rows()
        written_rows = 0

        def _iter() -> Iterator[pa.RecordBatch]:
            nonlocal written_rows
            while written_rows < num_rows:
                while not buffer or buffer[0][0] != written_rows:
                    batch: tuple[int, dict[str, str]] = self.checkpoint_keys.get()
                    heapq.heappush(buffer, batch)

                data = heapq.heappop(buffer)[1]
                res = pa.RecordBatch.from_pydict(
                    {
                        key: _combine_chunks(self.store[value]["data"])
                        for key, value in data.items()
                    }
                )
                yield res
                written_rows += res.num_rows

        it = more_itertools.peekable(_iter())
        pa.RecordBatchReader.from_batches(it.peek().schema, it)

        file_id = str(uuid.uuid4())

        with lance.file.LanceFileWriter(
            f"{self.uri}/data/{file_id}.lance",
            it.peek().schema,
        ) as writer:
            for batch in it:
                writer.write_batch(batch)

        # MASSIVE HACK: open up an API to get the field id from the column name
        field_id = re.compile(
            rf'name: "{self.column_name}", id: (?P<field_id>[\d]*),'
        ).findall(str(dataset.lance_schema))
        assert len(field_id) == 1
        field_id = int(field_id[0])

        new_datafile = lance.fragment.DataFile(
            f"{file_id}.lance",
            [field_id],
            [0],
            2,
            0,
        )

        new_datafile.fields = [field_id]
        new_datafile.column_indices = [0]

        return self.fragment_id, new_datafile


FragmentWriter: ray.actor.ActorClass = cast(ray.actor.ActorClass, FragmentWriter)


def run_ray_add_column(
    uri: str,
    columns: list[str],
    transforms: dict[str, UDF],
    checkpoint_store: CheckpointStore,
    /,
    job_id: str | None = None,
    batch_size: int = 8,
    read_version: int | None = None,
    concurrency: int = 8,
    applier_batch_size: int = 32,
    test_run: bool = True,
    **kwargs,
) -> None:
    if read_version is None:
        read_version = lance.dataset(uri).version

    job_id = job_id or uuid.uuid4().hex

    applier = LanceRecordBatchUDFApplier(
        udfs=transforms,
        checkpoint_store=checkpoint_store,
        error_logger=CheckpointStoreErrorLogger(job_id, checkpoint_store),
        batch_size=applier_batch_size,
    )

    pool = ray.util.ActorPool(
        [ApplierActor.remote(applier=applier) for _ in range(concurrency)]
    )

    plan = plan_read(
        uri,
        columns,
        batch_size=batch_size,
        read_version=read_version,
        **kwargs,
    )

    applier_iter = pool.map_unordered(
        lambda actor, value: actor.run.remote(value),
        # the API says list, but iterables are fine
        plan,
    )

    writers = {}
    writer_queues: dict[int, ray.util.queue.Queue] = {}
    writer_futs = {}

    applier_pbar = tqdm(applier_iter, total=len(plan), position=0)
    applier_pbar.set_description("Applying UDFs")

    writer_pbar = tqdm(total=len(writers), position=1)
    writer_pbar.set_description("Writing Fragments")

    def _commit(frags) -> None:
        nonlocal read_version
        _LOG.info("Committing %d fragments", len(ready))
        frags = ray.get(frags)
        operation = lance.LanceOperation.DataReplacement(
            replacements=[
                lance.LanceOperation.DataReplacementGroup(
                    fragment_id=frag_id,
                    new_file=new_file,
                )
                for frag_id, new_file in frags
            ]
        )
        lance.LanceDataset.commit(uri, operation, read_version=read_version)
        read_version += 1
        for frag_id, _ in frags:
            del writers[frag_id]
            del writer_queues[frag_id]
            del writer_futs[frag_id]
            writer_pbar.update(1)

    for item in applier_pbar:
        task: ReadTask = item[0]
        result = item[1]

        frag_id = task.frag_id
        if frag_id not in writers:
            _LOG.info("Creating writer for fragment %d", frag_id)
            queue = ray.util.queue.Queue()
            writer = FragmentWriter.remote(
                uri, list(transforms.keys())[0], checkpoint_store, frag_id, queue
            )
            writer_queues[frag_id] = queue
            writers[frag_id] = writer
            writer_futs[frag_id] = writer.write.remote()
        queue = writer_queues[frag_id]

        queue.put((task.offset, result))

        ready, _ = ray.wait(list(writer_futs.values()), timeout=0)
        if ready:
            _commit(ready)

    # commit any remaining fragments
    _commit(list(writer_futs.values()))
